#region using statements
using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using visLU.Effects;
using visLU.Gui;
#endregion

namespace visLU
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {

        #region variables
        GraphicsDeviceManager graphics;
        GraphicsDevice device;

        SpriteBatch spriteBatch;
        KeyboardState lastKeyboardState;

        MainWindow gui;
        IntPtr XNADrawSurface;

        Engine engine;

        int screenWidth;
        int screenHeight;
        #endregion

        public Game1()
        {
            
            //init gui window and get drawing surface handle
            gui = new MainWindow();
            gui.Show();
            gui.setXNADrawSurfaceDimensions();
            XNADrawSurface = gui.getDrawSurface();
            
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            //set event handler to draw to the picturebox control (XNADrawSurface)
            if (!GameProperties.Instance.fullscreen)
            {
                
                graphics.PreparingDeviceSettings += new EventHandler<PreparingDeviceSettingsEventArgs>(graphics_PreparingDeviceSettings);
                System.Windows.Forms.Control.FromHandle((this.Window.Handle)).VisibleChanged += new EventHandler(Game1_VisibleChanged);
                screenWidth = GameProperties.Instance.XNADrawSurfaceWidth;
                screenHeight = GameProperties.Instance.XNADrawSurfaceHeight;
                
            }

            //screenWidth = GameProperties.Instance.XNADrawSurfaceWidth;
            //screenHeight = GameProperties.Instance.XNADrawSurfaceHeight;
            
            // Antialiasing
            graphics.PreferMultiSampling = true;

            //init engine
            Camera[] cameras = new Camera[4];
            cameras[0] = new InteractiveCamera(this);
            base.Components.Add(cameras[0]);
            for (int i = 1; i < 4; i++)
            {
                cameras[i] = new Camera(this);
                base.Components.Add(cameras[i]);

            }
            this.Services.AddService(typeof(Camera[]), cameras);
            //Data data = new Data(GameProperties.Instance.dataDir + GameProperties.Instance.dataFilename, this);
            engine = new Engine(this);
            //engine.Input = data;
            base.Components.Add(engine);
            //base.Components.Add(data);
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            device = this.GraphicsDevice;
            setGraphicsDeviceManager();

            spriteBatch = new SpriteBatch(GraphicsDevice);
            Services.AddService(typeof(SpriteBatch), spriteBatch);

            base.Initialize();

        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            base.LoadContent();

        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            KeyboardState currentKeyboardState = Keyboard.GetState();
            
            // Allows the game to exit
            if (currentKeyboardState.IsKeyDown(Keys.Escape) || GameProperties.Instance.exitGame)
            {
                this.Exit();
                System.Windows.Forms.Application.Exit();
            }

            //update gui
            if (GameProperties.Instance.updateGuiForm)
            {
                gui.updateSliceControls(GameProperties.Instance.xDataRange, GameProperties.Instance.yDataRange, GameProperties.Instance.zDataRange);
                GameProperties.Instance.updateGuiForm = false;
            }
            lastKeyboardState = currentKeyboardState;
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            //GraphicsDevice.Clear(Color.Black);
            base.Draw(gameTime);
        }

        private void setGraphicsDeviceManager()
        {
            if (GameProperties.Instance.fullscreen)
            {
                graphics.PreferredBackBufferWidth = GraphicsDevice.DisplayMode.Width;
                graphics.PreferredBackBufferHeight = GraphicsDevice.DisplayMode.Height;
                graphics.ToggleFullScreen();
                graphics.IsFullScreen = true;
                GameProperties.Instance.fullscreenWidth = GraphicsDevice.DisplayMode.Width;
                GameProperties.Instance.fullscreenHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                graphics.PreferredBackBufferWidth = screenWidth;
                graphics.PreferredBackBufferHeight = screenHeight;
            }

            graphics.ApplyChanges();

            // Multi sampling
            graphics.PreferMultiSampling = true;

            // Minimum shader profile required
            graphics.MinimumVertexShaderProfile = ShaderProfile.VS_2_0;
            graphics.MinimumPixelShaderProfile = ShaderProfile.PS_2_0;

            //Max FPS:
            //this.IsFixedTimeStep = false;
            //graphics.SynchronizeWithVerticalRetrace = false;
        }
    

        #region event handler
        void graphics_PreparingDeviceSettings(object sender, PreparingDeviceSettingsEventArgs e)
        {
                e.GraphicsDeviceInformation.PresentationParameters.DeviceWindowHandle = XNADrawSurface;
           
        }
         private void Game1_VisibleChanged(object sender, EventArgs e)

        {

                if (System.Windows.Forms.Control.FromHandle((this.Window.Handle)).Visible == true)

                    System.Windows.Forms.Control.FromHandle((this.Window.Handle)).Visible = false;
        }

        #endregion
 
    }
}
